/***************************************************************************
 *   Copyright (C) 2005 by Aleš Zamuda   *
 *   ales.zamuda@uni-mb.si   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "geneticsimulation.h"
#include <qevent.h>
#include "communicationthread.h"
#include "transactionstartevent.h"
#include "evaluationparameters.h"
#include "populacija.h"
#include "osebek.h"
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <pthread.h>
#include "transactionstartevent.h"
#include <qapplication.h>
#include <sstream>

namespace genetic {

	GeneticSimulation::GeneticSimulation() {
		engine=0;
		targetWidget=0;
		isContinue=false;
		isMultiselection=false;
		populacija=new Populacija;
	}

	GeneticSimulation::~GeneticSimulation() {
	}

	void GeneticSimulation::setTargetWidget(QWidget* target) {
		targetWidget=target;
	}

	void GeneticSimulation::addTransaction(const std::string &s) {
		transactions.push_back(s);
	}

	GeneticSimulation* staticGeneticSimulation;
	
	void GeneticSimulation::start() {
		staticGeneticSimulation=this;
		pthread_t ThreadA;
		pthread_create(&ThreadA,NULL,threadExecute,NULL);
	}
	
	void* GeneticSimulation::threadExecute(void*) {
		staticGeneticSimulation->run();
		return 0;
	}
	
	void GeneticSimulation::run() {
		engine = new genetic::chessengine::EngineCommunication();
		engine->geneticSimulation=this;
		readFens();
		readFensResults();

		evaluationParameters = new chessengine::EvaluationParameters;
		engine->startEngineProcess();

		serveTransactions();
	}

	void GeneticSimulation::readFens() {
		std::ifstream in("../../engine/data/fens.txt");
		string line;
		while (in.good()) {
			getline(in, line);
			engine->fens.push_back(line);
		}
	}
	
	void GeneticSimulation::readFensResults() {
		std::ifstream in("../../engine/data/fensResults.txt");
		string line;
		while (in.good()) {
			getline(in, line);
			engine->results.push_back(line);
		}
	}

	void GeneticSimulation::posljiRefreshMsgVGui() {
		gui::TransactionStartEvent *evt = new gui::TransactionStartEvent;
		evt->message = "refreshStats";
		QApplication::postEvent(targetWidget, evt);
	}

	void GeneticSimulation::posljiEvalMsgVGui(const std::string &s) {
		gui::TransactionStartEvent *evt = new gui::TransactionStartEvent;
		evt->message = s;
		QApplication::postEvent(targetWidget, evt);
	}

	void GeneticSimulation::posljiFinishMsgVGui() {
		gui::TransactionStartEvent *evt = new gui::TransactionStartEvent;
		evt->message = "finish";
		QApplication::postEvent(targetWidget, evt);
	}

	void GeneticSimulation::serveTransactions() {
		for (;;) {
			if ((int)transactions.size()>0) {
				if (!isMultiselection) {//normalni zagon
					if (!isContinue) doInitPopulation();
	
					while (populacija->naslednjaGeneracija()) {
						saveBestMessage();
						posljiRefreshMsgVGui();
					}
					posljiRefreshMsgVGui();
				} else {
					doInitPopulation();

					std::vector<Osebek*> *initOsebki = populacija->getOsebkiCopy();
					assert(initOsebki);
					
					for (int i = 0; i < 5; i++) {
						populacija->resetStGeneracije();
						populacija->setInitOsebki(initOsebki);
						populacija->changeSelectionType(
								i==0?Populacija::SELECTION_PROPORTIONAL:
								i==1?Populacija::SELECTION_BAKER:
								i==2?Populacija::SELECTION_LINEAR:
								i==3?Populacija::SELECTION_NOLINEAR:
								Populacija::SELECTION_TOURNAMENT);
		
						while (populacija->naslednjaGeneracija()) {
							saveBestMessage();
							posljiRefreshMsgVGui();
						}
					}
					posljiRefreshMsgVGui();
					for (int i = 0; i < populacija->getPop_size(); i++) delete (*initOsebki)[i];
					delete initOsebki;

					//izpis stats:
					populacija->printSelectionStats();
				}
				posljiFinishMsgVGui();
				transactions.pop_front();
			}
			usleep(100000);
		}
	}
	
	//napolnimo nekaj privzetih genov
	void GeneticSimulation::doInitPopulation() {
		populacija->setM(evaluationParameters->getNrBitsNeededForGenes());
		populacija->geneticSimulation = this;
		populacija->init_pop();

		//inicializiramo prvi osebek z default kromosomom
		std::vector<bool> defaultChromosome;
		evaluationParameters->getDefaultChromosome(defaultChromosome);
		populacija->setDefaultChromosome(defaultChromosome);
	}

	float GeneticSimulation::evaluate(Osebek *osebek) {
		evaluationParameters->fromBits(osebek->biti);
		
		std::cerr << "evaluating #";
		int id = osebek->id;
		std::cerr << id << " = (" << evaluationParameters->serializeParameters()<<")"<<std::flush;
		float ocena=engine->oceniGene(evaluationParameters->serializeParameters());
		std::cerr<<" -> " << ocena << std::endl;
		return ocena;
	}
	
	void GeneticSimulation::saveBestMessage() {
		evaluationParameters->fromBits(populacija->getBestChromosomes());
		bestMessage = evaluationParameters->serializeParameters();
		static std::ofstream out("ocene.txt");//hranimo zaradi varnosti
		out << bestMessage << std::endl;
	}

	/**
	 * Iz datoteke prebere za pop_size genov in inicializira populacijo.
	 * @param fileName 
	 */
	void GeneticSimulation::loadFile(const std::string &fileName) {
		std::ifstream vhod(fileName.c_str());
		
		assert(vhod.good());
		std::cerr<<"ODPRL:"<<fileName<<std::endl;
		int G, stG, pop_size;
		float p_k, p_m;
		int m, e;
		
		vhod >> G;
		populacija->setG(G);
		vhod >> stG;
		populacija->setStGeneracije(stG);
		vhod >> pop_size;
		populacija->setPop_size(pop_size);
		vhod >> p_k;
		populacija->setP_k(p_k);
		vhod >> p_m;
		populacija->setP_m(p_m);
		vhod >> m;
		populacija->setM(m);
		vhod >> e;
		populacija->setE(e);

		doInitPopulation();
		isContinue=true;

		std::vector<Osebek*> *initOsebki = populacija->getOsebkiCopy();
		assert(initOsebki);
		
		for (int i = 0; i < pop_size; i++) {
			evaluationParameters->readChromosome(vhod);
			evaluationParameters->toBits((*initOsebki)[i]->biti);
		}

		populacija->setInitOsebki(initOsebki);

		for (int i = 0; i < populacija->getPop_size(); i++) delete (*initOsebki)[i];
			delete initOsebki;
		posljiRefreshMsgVGui();
	}

	void GeneticSimulation::saveFile( const std::string & fileName ) {
		std::ofstream izhod(fileName.c_str());
		assert(izhod.good());

		izhod << populacija->getG() << " ";
		izhod << populacija->getStGeneracije()<< " ";
		izhod << populacija->getPop_size()<< " ";
		izhod << populacija->getP_k()<< " ";
		izhod << populacija->getP_m()<< " ";
		izhod << populacija->getM()<< " ";
		izhod << populacija->getE()<< std::endl;

		std::vector<Osebek*> *initOsebki = populacija->getOsebkiCopy();
		assert(initOsebki);
		int pop_size = populacija->getPop_size();

		for (int i = 0; i < pop_size; i++) {
			evaluationParameters->fromBits((*initOsebki)[i]->biti);
			string s = evaluationParameters->serializeParameters();
			izhod << s<<std::endl;
		}

		for (int i = 0; i < pop_size; i++) delete (*initOsebki)[i];
		delete initOsebki;
	}
}

